#ifndef VECTOR_HPP
#define VECTOR_HPP

#include <algorithm>
#include <cassert> 

template<
    typename T,
    typename checking_policy = no_check_policy
>
class Vector;

struct resize_policy;
struct assert_policy;
struct no_check_policy;


template<
    typename T,
    typename checking_policy = no_check_policy
>
class Vector
{
public:
    Vector() = default;
    Vector(size_t size);
    Vector(const Vector&);
    Vector(Vector&&);

    Vector& operator=(const Vector&);
    Vector& operator=(Vector&&);

    void push_back(const T&);
    T pop_back();

    T* begin();
    const T* begin() const;

    T* end();
    const T* end() const;

    T& operator[](size_t at);
    const T& operator[](size_t at) const;

    void resize(size_t size);

    size_t count() const { return count_; }
    size_t size() const { return size_; }

    ~Vector();
private:
    void swap(Vector<T, checking_policy>& other);

    bool is_full();

    T* data_ = nullptr;
    size_t size_ = 0;
    size_t count_ = 0;

};

template<typename T, typename checking_policy>
inline Vector<T, checking_policy>::Vector(size_t size)
    : size_(size), count_(0), data_(new T[size])
{
}

template<typename T, typename checking_policy>
inline Vector<T, checking_policy>::Vector(const Vector<T, checking_policy> & other)
    : Vector<T, checking_policy>(other.size_)
{
    for (auto& val : other)
        push_back(val);
}

template<typename T, typename checking_policy>
inline Vector<T, checking_policy>::Vector(Vector && other)
    : Vector<T>(0)
{
    swap(other);
}

template<typename T, typename checking_policy>
inline Vector<T, checking_policy> & Vector<T, checking_policy>::operator=(const Vector<T, checking_policy> & other)
{
    return *this = Vector<T, checking_policy>(other);
}

template<typename T, typename checking_policy>
inline Vector<T, checking_policy> & Vector<T, checking_policy>::operator=(Vector<T, checking_policy> && other)
{
    swap(other);
    return *this;
}

template<typename T, typename checking_policy>
inline void Vector<T, checking_policy>::push_back(const T & value)
{
    if (is_full())
    {
        resize(size_ * 2 ? size_ * 2 : 1);
    }
    data_[count_++] = value;
}

template<typename T, typename checking_policy>
inline T Vector<T, checking_policy>::pop_back()
{
    return data_[--count_];
}

template<typename T, typename checking_policy>
inline T * Vector<T, checking_policy>::begin()
{
    return data_;
}

template<typename T, typename checking_policy>
inline const T * Vector<T, checking_policy>::begin() const
{
    return data_;
}

template<typename T, typename checking_policy>
inline T * Vector<T, checking_policy>::end()
{
    return data_ + count_;
}

template<typename T, typename checking_policy>
inline const T * Vector<T, checking_policy>::end() const
{
    return data_ + count_;
}

template<typename T, typename checking_policy>
inline T & Vector<T, checking_policy>::operator[](size_t at)
{
    checking_policy::check(*this, at);
    return data_[at];
}

template<typename T, typename checking_policy>
inline const T & Vector<T, checking_policy>::operator[](size_t at) const
{
    checking_policy::check(count_, at);
    return data_[at];
}

template<typename T, typename checking_policy>
inline Vector<T, checking_policy>::~Vector()
{
    delete[] data_;
    size_ = 0;
    count_ = 0;
}

template<typename T, typename checking_policy>
inline void Vector<T, checking_policy>::swap(Vector<T, checking_policy>& other)
{
    std::swap(data_, other.data_);
    std::swap(size_, other.size_);
    std::swap(count_, other.count_);
}

template<typename T, typename checking_policy>
inline bool Vector<T, checking_policy>::is_full()
{
    return size_ == count_;
}

template<typename T, typename checking_policy>
inline void Vector<T, checking_policy>::resize(size_t size)
{
    auto new_vector = Vector<T, checking_policy>(size);

    for (auto& val : *this)
        new_vector.push_back(val);

    *this = new_vector;
}

struct no_check_policy
{
    template<typename T, typename checking_policy = no_check_policy>
    static void check(Vector<T, checking_policy>& vector, size_t at)
    {
    }
};

struct assert_policy
{
    template<typename T, typename checking_policy>
    static void check(Vector<T, checking_policy>& vector, size_t at)
    {
        assert(vector.count() > at);
    }
};

struct resize_policy
{
    template<typename T, typename checking_policy>
    static void check(Vector<T, checking_policy>& vector, size_t at)
    {
        if (at > vector.size())
        {
            while (vector.count() <= at)
            {
                vector.push_back(T());
            }
        }
    }
};


#endif // VECTOR_HPP

